@@ -42,8 +42,7 @@
   :group 'epg
   :type '(choice (const :tag "Default" nil) directory))
 
-(defcustom epg-passphrase-coding-system (if (boundp 'locale-coding-system)
-					    locale-coding-system)
+(defcustom epg-passphrase-coding-system nil
   "Coding system to use with messages from `epg-gpg-program'."
   :group 'epg
   :type 'symbol)
@@ -54,7 +53,7 @@ Note that the buffer name starts with a space."
   :group 'epg
   :type 'boolean)
 
-(defconst epg-version-number "0.0.6")
+(defconst epg-version-number "0.0.9")
 
 (defconst epg-gpg-minimum-version "1.4.3")
 
